#include <stdio.h>

#if TARGET_DOS

  #include <dos.h>

  #define Outb(prt, val)   outportb(prt, val)
  #define Inpb(prt)        inportb(prt)

#elif TARGET_WINXP

  #include <conio.h>
  #include <windows.h>

  #define delay(x)	    Sleep(x)
  #define Outb(prt, val)    Out32(prt, val)
  #define Inpb(prt)         Inp32(prt)

#else /* TARGET LINUX */

#include <stdlib.h>
#include <unistd.h>
#include <asm/io.h>

#define delay(x)            usleep((x)*1000)
#define Outb(prt, val)      outb((val), (prt))
#define Inpb(prt)           inb((prt))

#endif

#ifndef TRUE
#define TRUE 1
#endif

#ifndef FALSE
#define FALSE 0
#endif


/* counter states */

#define INI     0
#define FW1     1
#define FW2     2
#define FW3     3
#define BW1     4
#define BW2     5
#define BW3     6
#define DECR    7
#define INCR    8
#define EXIT    9

int counter;
int state;


#if TARGET_DOS

int beep_on;

#endif

#if TARGET_WINXP

typedef short (STDAPICALLTYPE *inpfuncPtr)(short portaddr);
typedef void (STDAPICALLTYPE *oupfuncPtr)(short portaddr, short datum);
static  inpfuncPtr inp32fp;
static  oupfuncPtr oup32fp;

static short Inp32(short portaddr)
{
    return (inp32fp)(portaddr);
}

static void  Out32(short portaddr, short datum)
{
     (oup32fp)(portaddr, datum);
} 

#endif

int no_beep = FALSE;

static int reset_to_initial_position();

static int get_new_position(int current_pos);

static void print_new_counter_value();

static void usage(void)
{
#if TARGET_DOS || TARGET_WINXP
    printf("\nUsage:\n"
           "counter.exe [ddd] [-no_beep] [-help]\n"
           "ddd is an initial counter value (decimal), default is 0\n"
           "-no_beep  - turns off beep sound\n"
           "-help  - print this message\n");
#else
    printf("\nUsage:\n"
           "counter.exe [ddd] [-help]\n"
           "ddd is an initial counter value (decimal), default is 0\n"
           "-help  - print this message\n");
#endif

}

int
main(int argc, char** argv)
{

    int arg;
#if TARGET_WINXP
     HINSTANCE hLib;
#endif

    for (arg = 1; arg < argc; arg++) {
        char* aptr = argv[arg];

        if (isdigit(*aptr)) {
           sscanf(aptr, "%d", &counter);
        }
        else if (strcmp(aptr, "-help") == 0) {
           usage();
           return 0;
        }
        else if (strcmp(aptr, "-no_beep") == 0) {
           no_beep = TRUE;
	   printf("Beep is off\n");
        }
        else {
           printf("Unrecognized option %s\n", aptr);
	   usage();
           return;
        }

    }

#if TARGET_WINXP
    /* Load the library */
    hLib = LoadLibrary("inpout32.dll");

    if (hLib == NULL) {
        fprintf(stderr,"inpout32.dll library load failed.\n");
        return 1;
    }

    /* get the address of the function */

    inp32fp = (inpfuncPtr) GetProcAddress(hLib, "Inp32");

    if (inp32fp == NULL) {
        fprintf(stderr,"GetProcAddress for Inp32 Failed.\n");
        return 1;
    }


    oup32fp = (oupfuncPtr) GetProcAddress(hLib, "Out32");

    if (oup32fp == NULL) {
        fprintf(stderr,"GetProcAddress for Out32 Failed.\n");
        return 1;
    }

#endif

    print_new_counter_value();
    printf("\n");

    state = reset_to_initial_position();

    if (state != INI) {
        return 0;
    }
    
    do {
        state = Counter_get_next_state(state);
        if (state == INCR) {
            counter++;
            print_new_counter_value();
        }
        else if (state == DECR) {
            counter--;
            print_new_counter_value();
        }
    } while (state != EXIT);

    printf ("\n");

#if TARGET_WINXP
    FreeLibrary(hLib);
#endif

    return 0;
}

int
Counter_get_next_state(int current_state)
{
    int position;

    switch (current_state) {
        case INI:
        case INCR:
        case DECR:
            position = get_new_position(3);
            if (position == 1) {
                return BW1;
            }
            else if (position == 2) {
                return FW1;
            }
            break;
        case FW1:
            position = get_new_position(2);
            if (position == 3) {
                return INI;
            }
            else if (position == 0) {
                return FW2;
            }
            break;
        case FW2:
            position = get_new_position(0);
            if (position == 2) {
                return FW1;
            }
            else if (position == 1) {
                return FW3;
            }
            break;
        case FW3:
            position = get_new_position(1);
            if (position == 0) {
                return FW2;
            }
            else if (position == 3) {
                return INCR;
            }
            break;
        case BW1:
            position = get_new_position(1);
            if (position == 3) {
                return INI;
            }
            else if (position == 0) {
                return BW2;
            }
            break;
        case BW2:
            position = get_new_position(0);
            if (position == 1) {
                return BW1;
            }
            else if (position == 2) {
                return BW3;
            }
            break;
        case BW3:
            position = get_new_position(2);
            if (position == 0) {
                return BW2;
            }
            else if (position == 3) {
                return DECR;
            }
            break;
        default:
            break;
    }
    if(position != 4) {
      printf ( "Error: got unexpected position %d after %d\n", position, current_state);
#if TARGET_DOS
    if (no_beep != TRUE) {
      sound(200);
      delay(800);
      nosound();
    }
#elif TARGET_WINXP
    if (no_beep != TRUE) {
      Beep(200, 800);
    }
#endif
    }

    return EXIT;
}



/*
 *
 * IBM-PC Parallel Printer Port Data & Status Registers
 * ====================================================
 * I/O Port
 * Data    Base = 278/378/3BC Hex
 * Status  Base + 1  (bit S7 is inverted)
 */

#define PData    0x378
#define PStatus  PData + 1

static int
reset_to_initial_position()
{
    int value;

    printf("\n");
    printf("Parallel Port counter\n");
    printf("===================================\n");
    printf("\n");

#if TARGET_LINUX

    if(ioperm(PData, 2, 1)) {
        printf("Couldn't get the port at 0x%x\n, PData");
        return EXIT;
    }
#endif


    /*
     * check the presense of counter device.
     *
     * D0 is connected with S5 in the device
     * S5 should read the value we set to D0
     * S6 should read 1. that means that power is on.
     * We also check that S7 is 0.
     */

    /* First toggle D0 and read S5 to check the connection */
    Outb(PData,1); 
    delay(200);
    value = Inpb(PStatus);
	
    value = (value >> 5) & 1;
    if (value != 1) {
        printf("LPT counter device not found.\n");
        printf("Check that device is connected and power is on and then try again\n");
        return EXIT;
    }


    Outb(PData,0); 
    delay(200);
    value = Inpb(PStatus);
  
    value = (value >> 5) & 0x7;	

    if (value & 1) {
        printf("LPT counter device not found.\n");
        printf("Check that device is connected and power is on and then try again\n");
        return EXIT;
    }

#if DEBUG
    printf("0x%x\n", value); 
#endif

    /* Check S7 */
    if (value & 0x4) {
        printf("Exit button is pushed.\n");
        return EXIT;
    }
    /* Check S6 */
    else if (value != 0x2) {
        printf("Check the device power.\n");
        return EXIT;
    }

    do {
        value = Inpb(PStatus);

        if (value & 0x80) {
            /* pressed exit button */
            return EXIT;
        }

        value = (value >> 3) & 0x3;

        /* we are at know state when both sensors are highlighted */
    } while (value != 3);

    return INI;
}

static int
get_new_position(int current_pos)
{
    /* 
     * Returns new sensors position when position 
     * changes relative to the current position.
     * Position is an integer value indicating one 
     * the following:
     *
     * return value   |  HL1  |  HL2
     * -----------------------------
     *              0 |  off  |  off
     *              1 |  on   |  off
     *              2 |  off  |  on
     *              3 |  on   |  on
     *              4 |  exit
     * -----------------------------
     */
    int value;
    do {

#if TARGET_DOS
        delay(5);

        if ((no_beep != TRUE) && (beep_on != 0)) {
	    beep_on += 5;
            if (beep_on > 50) {
                nosound();
                beep_on = 0;
            }
        }
#endif
        value = Inpb(PStatus);

        if (value & 0x80) {
            /* pressed exit button */
            return 4;
        }
        /* S3 is HL1 and S4 is HL2 */
        value = (value >> 3) & 0x3;
    } while (value == current_pos);

    return value;
}

static void
print_new_counter_value()
{

    printf("W= %-10d\r", counter);

#if TARGET_DOS
    if (no_beep != TRUE) {
        if (state == INCR) {
            sound(800);
            beep_on = 1;
        } else if (state == DECR) {
            sound(600);
            beep_on = 1;
        }
    }
#elif TARGET_WINXP
    if (no_beep != TRUE) {
        if (state == INCR) {
            Beep(800, 30);
        } else if (state == DECR) {
            Beep(600, 30);
        }
    }
#endif

}
